@@ -12,21 +12,43 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-"""OpTree: Optimized PyTree Utilities."""
+"""Registry for custom pytree node types."""
 
 from __future__ import annotations
 
+import contextlib
 import dataclasses
 import functools
 import inspect
 import sys
+import warnings
 from collections import OrderedDict, defaultdict, deque, namedtuple
-from operator import methodcaller
+from operator import itemgetter, methodcaller
 from threading import Lock
-from typing import TYPE_CHECKING, Any, Callable, Iterable, NamedTuple, Sequence, overload
-from typing_extensions import Self  # Python 3.11+
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    ClassVar,
+    Generator,
+    Iterable,
+    NamedTuple,
+    Sequence,
+    Type,
+    overload,
+)
+from typing_extensions import TypeAlias  # Python 3.10+
+from typing_extensions import deprecated  # Python 3.13+
 
 from optree import _C
+from optree.accessor import (
+    AutoEntry,
+    MappingEntry,
+    NamedTupleEntry,
+    PyTreeEntry,
+    SequenceEntry,
+    StructSequenceEntry,
+)
 from optree.typing import (
     KT,
     VT,
@@ -50,21 +72,25 @@ __all__ = [
     'register_pytree_node',
     'register_pytree_node_class',
     'unregister_pytree_node',
-    'Partial',
-    'register_keypaths',
-    'AttributeKeyPathEntry',
-    'GetitemKeyPathEntry',
+    'dict_insertion_ordered',
 ]
 
 
-SLOTS = {'slots': True} if sys.version_info >= (3, 10) else {}
+SLOTS = {'slots': True} if sys.version_info >= (3, 10) else {}  # Python 3.10+
 
 
 @dataclasses.dataclass(init=True, repr=True, eq=True, frozen=True, **SLOTS)
 class PyTreeNodeRegistryEntry:
+    """A dataclass that stores the information of a pytree node type."""
+
     type: builtins.type
     flatten_func: FlattenFunc
     unflatten_func: UnflattenFunc
+
+    if sys.version_info >= (3, 10):
+        _: dataclasses.KW_ONLY  # Python 3.10+
+
+    path_entry_type: builtins.type[PyTreeEntry] = AutoEntry
     namespace: str = ''
 
 
@@ -73,6 +99,8 @@ del SLOTS
 
 # pylint: disable-next=missing-class-docstring,too-few-public-methods
 class GlobalNamespace:  # pragma: no cover
+    __slots__: ClassVar[tuple[()]] = ()
+
     def __repr__(self) -> str:
         return '<GLOBAL NAMESPACE>'
 
@@ -82,12 +110,17 @@ __REGISTRY_LOCK: Lock = Lock()
 del GlobalNamespace
 
 
+CustomTreeNodeT: TypeAlias = Type[CustomTreeNode[T]]
+
+
 def register_pytree_node(
-    cls: type[CustomTreeNode[T]],
+    cls: CustomTreeNodeT,
     flatten_func: FlattenFunc,
     unflatten_func: UnflattenFunc,
+    *,
+    path_entry_type: type[PyTreeEntry] = AutoEntry,
     namespace: str,
-) -> type[CustomTreeNode[T]]:
+) -> CustomTreeNodeT:
     """Extend the set of types that are considered internal nodes in pytrees.
 
     See also :func:`register_pytree_node_class` and :func:`unregister_pytree_node`.
@@ -113,6 +146,8 @@ def register_pytree_node(
         unflatten_func (callable): A function taking two arguments: the auxiliary data that was
             returned by ``flatten_func`` and stored in the treespec, and the unflattened children.
             The function should return an instance of ``cls``.
+        path_entry_type (type, optional): The type of the path entry to be used in the treespec.
+            (default: :class:`AutoEntry`)
         namespace (str): A non-empty string that uniquely identifies the namespace of the type registry.
             This is used to isolate the registry from other modules that might register a different
             custom behavior for the same type.
@@ -122,6 +157,7 @@ def register_pytree_node(
 
     Raises:
         TypeError: If the input type is not a class.
+        TypeError: If the path entry class is not a subclass of :class:`PyTreeEntry`.
         TypeError: If the namespace is not a string.
         ValueError: If the namespace is an empty string.
         ValueError: If the type is already registered in the registry.
@@ -203,9 +239,11 @@ def register_pytree_node(
         )
     """
     if not inspect.isclass(cls):
-        raise TypeError(f'Expected a class, got {cls}.')
+        raise TypeError(f'Expected a class, got {cls!r}.')
+    if not (inspect.isclass(path_entry_type) and issubclass(path_entry_type, PyTreeEntry)):
+        raise TypeError(f'Expected a subclass of PyTreeEntry, got {path_entry_type!r}.')
     if namespace is not __GLOBAL_NAMESPACE and not isinstance(namespace, str):
-        raise TypeError(f'The namespace must be a string, got {namespace}.')
+        raise TypeError(f'The namespace must be a string, got {namespace!r}.')
     if namespace == '':
         raise ValueError('The namespace cannot be an empty string.')
 
@@ -217,11 +255,18 @@ def register_pytree_node(
         registration_key = (namespace, cls)
 
     with __REGISTRY_LOCK:
-        _C.register_node(cls, flatten_func, unflatten_func, namespace)
+        _C.register_node(
+            cls,
+            flatten_func,
+            unflatten_func,
+            path_entry_type,
+            namespace,
+        )
         _NODETYPE_REGISTRY[registration_key] = PyTreeNodeRegistryEntry(
             cls,
             flatten_func,
             unflatten_func,
+            path_entry_type=path_entry_type,
             namespace=namespace,
         )
     return cls
@@ -231,25 +276,26 @@ def register_pytree_node(
 def register_pytree_node_class(
     cls: str | None = None,
     *,
+    path_entry_type: type[PyTreeEntry] | None = None,
     namespace: str | None = None,
-) -> Callable[[type[CustomTreeNode[T]]], type[CustomTreeNode[T]]]:  # pragma: no cover
-    ...
+) -> Callable[[CustomTreeNodeT], CustomTreeNodeT]: ...
 
 
 @overload
 def register_pytree_node_class(
-    cls: type[CustomTreeNode[T]],
+    cls: CustomTreeNodeT,
     *,
+    path_entry_type: type[PyTreeEntry] | None,
     namespace: str,
-) -> type[CustomTreeNode[T]]:  # pragma: no cover
-    ...
+) -> CustomTreeNodeT: ...
 
 
-def register_pytree_node_class(
-    cls: type[CustomTreeNode[T]] | str | None = None,
+def register_pytree_node_class(  # noqa: C901
+    cls: CustomTreeNodeT | str | None = None,
     *,
+    path_entry_type: type[PyTreeEntry] | None = None,
     namespace: str | None = None,
-) -> type[CustomTreeNode[T]] | Callable[[type[CustomTreeNode[T]]], type[CustomTreeNode[T]]]:
+) -> CustomTreeNodeT | Callable[[CustomTreeNodeT], CustomTreeNodeT]:
     """Extend the set of types that are considered internal nodes in pytrees.
 
     See also :func:`register_pytree_node` and :func:`unregister_pytree_node`.
@@ -266,6 +312,8 @@ def register_pytree_node_class(
 
     Args:
         cls (type, optional): A Python type to treat as an internal pytree node.
+        path_entry_type (type, optional): The type of the path entry to be used in the treespec.
+            (default: :class:`AutoEntry`)
         namespace (str, optional): A non-empty string that uniquely identifies the namespace of the
             type registry. This is used to isolate the registry from other modules that might
             register a different custom behavior for the same type.
@@ -275,6 +323,7 @@ def register_pytree_node_class(
         function that registers the class as a pytree node.
 
     Raises:
+        TypeError: If the path entry class is not a subclass of :class:`PyTreeEntry`.
         TypeError: If the namespace is not a string.
         ValueError: If the namespace is an empty string.
         ValueError: If the type is already registered in the registry.
@@ -284,12 +333,14 @@ def register_pytree_node_class(
 
         @register_pytree_node_class(namespace='foo')
         class Special:
+            TREE_PATH_ENTRY_TYPE = GetAttrEntry
+
             def __init__(self, x, y):
                 self.x = x
                 self.y = y
 
             def tree_flatten(self):
-                return ((self.x, self.y), None)
+                return ((self.x, self.y), None, ('x', 'y'))
 
             @classmethod
             def tree_unflatten(cls, metadata, children):
@@ -297,6 +348,8 @@ def register_pytree_node_class(
 
         @register_pytree_node_class('mylist')
         class MyList(UserList):
+            TREE_PATH_ENTRY_TYPE = SequenceEntry
+
             def tree_flatten(self):
                 return self.data, None, None
 
@@ -309,20 +362,38 @@ def register_pytree_node_class(
             raise ValueError('Cannot specify `namespace` when the first argument is a string.')
         if cls == '':
             raise ValueError('The namespace cannot be an empty string.')
-        return functools.partial(register_pytree_node_class, namespace=cls)  # type: ignore[return-value]
+        return functools.partial(
+            register_pytree_node_class,
+            path_entry_type=path_entry_type,
+            namespace=cls,
+        )  # type: ignore[return-value]
 
     if namespace is None:
         raise ValueError('Must specify `namespace` when the first argument is a class.')
     if namespace is not __GLOBAL_NAMESPACE and not isinstance(namespace, str):
-        raise TypeError(f'The namespace must be a string, got {namespace}')
+        raise TypeError(f'The namespace must be a string, got {namespace!r}')
     if namespace == '':
         raise ValueError('The namespace cannot be an empty string.')
 
     if cls is None:
-        return functools.partial(register_pytree_node_class, namespace=namespace)  # type: ignore[return-value]
+        return functools.partial(
+            register_pytree_node_class,
+            path_entry_type=path_entry_type,
+            namespace=namespace,
+        )  # type: ignore[return-value]
     if not inspect.isclass(cls):
-        raise TypeError(f'Expected a class, got {cls}.')
-    register_pytree_node(cls, methodcaller('tree_flatten'), cls.tree_unflatten, namespace)
+        raise TypeError(f'Expected a class, got {cls!r}.')
+    if path_entry_type is None:
+        path_entry_type = getattr(cls, 'TREE_PATH_ENTRY_TYPE', AutoEntry)
+    if not (inspect.isclass(path_entry_type) and issubclass(path_entry_type, PyTreeEntry)):
+        raise TypeError(f'Expected a subclass of PyTreeEntry, got {path_entry_type!r}.')
+    register_pytree_node(
+        cls,
+        methodcaller('tree_flatten'),
+        cls.tree_unflatten,
+        path_entry_type=path_entry_type,
+        namespace=namespace,
+    )
     return cls
 
 
@@ -365,9 +436,9 @@ def unregister_pytree_node(
         >>> unregister_pytree_node(set, namespace='temp')
     """
     if not inspect.isclass(cls):
-        raise TypeError(f'Expected a class, got {cls}.')
+        raise TypeError(f'Expected a class, got {cls!r}.')
     if namespace is not __GLOBAL_NAMESPACE and not isinstance(namespace, str):
-        raise TypeError(f'The namespace must be a string, got {namespace}.')
+        raise TypeError(f'The namespace must be a string, got {namespace!r}.')
     if namespace == '':
         raise ValueError('The namespace cannot be an empty string.')
 
@@ -384,11 +455,7 @@ def unregister_pytree_node(
 
 
 def _sorted_items(items: Iterable[tuple[KT, VT]]) -> list[tuple[KT, VT]]:
-    return total_order_sorted(items, key=lambda kv: kv[0])
-
-
-def _sorted_keys(dct: dict[KT, VT]) -> list[KT]:
-    return total_order_sorted(dct)
+    return total_order_sorted(items, key=itemgetter(0))
 
 
 def _none_flatten(none: None) -> tuple[tuple[()], None]:
@@ -427,6 +494,17 @@ def _dict_unflatten(keys: list[KT], values: Iterable[VT]) -> dict[KT, VT]:
     return dict(safe_zip(keys, values))
 
 
+def _dict_insertion_ordered_flatten(
+    dct: dict[KT, VT],
+) -> tuple[tuple[VT, ...], list[KT], tuple[KT, ...]]:
+    keys, values = unzip2(dct.items())
+    return values, list(keys), keys
+
+
+def _dict_insertion_ordered_unflatten(keys: list[KT], values: Iterable[VT]) -> dict[KT, VT]:
+    return dict(safe_zip(keys, values))
+
+
 def _ordereddict_flatten(
     dct: OrderedDict[KT, VT],
 ) -> tuple[tuple[VT, ...], list[KT], tuple[KT, ...]]:
@@ -442,7 +520,7 @@ def _defaultdict_flatten(
     dct: defaultdict[KT, VT],
 ) -> tuple[tuple[VT, ...], tuple[Callable[[], VT] | None, list[KT]], tuple[KT, ...]]:
     values, keys, entries = _dict_flatten(dct)
-    return values, (dct.default_factory, list(keys)), entries
+    return values, (dct.default_factory, keys), entries
 
 
 def _defaultdict_unflatten(
@@ -450,7 +528,22 @@ def _defaultdict_unflatten(
     values: Iterable[VT],
 ) -> defaultdict[KT, VT]:
     default_factory, keys = metadata
-    return defaultdict(default_factory, safe_zip(keys, values))
+    return defaultdict(default_factory, _dict_unflatten(keys, values))
+
+
+def _defaultdict_insertion_ordered_flatten(
+    dct: defaultdict[KT, VT],
+) -> tuple[tuple[VT, ...], tuple[Callable[[], VT] | None, list[KT]], tuple[KT, ...]]:
+    values, keys, entries = _dict_insertion_ordered_flatten(dct)
+    return values, (dct.default_factory, keys), entries
+
+
+def _defaultdict_insertion_ordered_unflatten(
+    metadata: tuple[Callable[[], VT], list[KT]],
+    values: Iterable[VT],
+) -> defaultdict[KT, VT]:
+    default_factory, keys = metadata
+    return defaultdict(default_factory, _dict_insertion_ordered_unflatten(keys, values))
 
 
 def _deque_flatten(deq: deque[T]) -> tuple[deque[T], int | None]:
@@ -479,15 +572,15 @@ def _structseq_unflatten(cls: type[structseq[T]], children: Iterable[T]) -> stru
 
 # pylint: disable=all
 _NODETYPE_REGISTRY: dict[type | tuple[str, type], PyTreeNodeRegistryEntry] = {  # fmt: off
-    type(None): PyTreeNodeRegistryEntry(type(None), _none_flatten, _none_unflatten),  # type: ignore[arg-type]
-    tuple: PyTreeNodeRegistryEntry(tuple, _tuple_flatten, _tuple_unflatten),  # type: ignore[arg-type]
-    list: PyTreeNodeRegistryEntry(list, _list_flatten, _list_unflatten),  # type: ignore[arg-type]
-    dict: PyTreeNodeRegistryEntry(dict, _dict_flatten, _dict_unflatten),  # type: ignore[arg-type]
-    namedtuple: PyTreeNodeRegistryEntry(namedtuple, _namedtuple_flatten, _namedtuple_unflatten),  # type: ignore[arg-type,dict-item] # noqa: PYI024
-    OrderedDict: PyTreeNodeRegistryEntry(OrderedDict, _ordereddict_flatten, _ordereddict_unflatten),  # type: ignore[arg-type]
-    defaultdict: PyTreeNodeRegistryEntry(defaultdict, _defaultdict_flatten, _defaultdict_unflatten),  # type: ignore[arg-type]
-    deque: PyTreeNodeRegistryEntry(deque, _deque_flatten, _deque_unflatten),  # type: ignore[arg-type]
-    structseq: PyTreeNodeRegistryEntry(structseq, _structseq_flatten, _structseq_unflatten),  # type: ignore[arg-type]
+    type(None): PyTreeNodeRegistryEntry(type(None), _none_flatten, _none_unflatten, path_entry_type=PyTreeEntry),  # type: ignore[arg-type]
+    tuple: PyTreeNodeRegistryEntry(tuple, _tuple_flatten, _tuple_unflatten, path_entry_type=SequenceEntry),  # type: ignore[arg-type]
+    list: PyTreeNodeRegistryEntry(list, _list_flatten, _list_unflatten, path_entry_type=SequenceEntry),  # type: ignore[arg-type]
+    dict: PyTreeNodeRegistryEntry(dict, _dict_flatten, _dict_unflatten, path_entry_type=MappingEntry),  # type: ignore[arg-type]
+    namedtuple: PyTreeNodeRegistryEntry(namedtuple, _namedtuple_flatten, _namedtuple_unflatten, path_entry_type=NamedTupleEntry),  # type: ignore[arg-type,dict-item] # noqa: PYI024
+    OrderedDict: PyTreeNodeRegistryEntry(OrderedDict, _ordereddict_flatten, _ordereddict_unflatten, path_entry_type=MappingEntry),  # type: ignore[arg-type]
+    defaultdict: PyTreeNodeRegistryEntry(defaultdict, _defaultdict_flatten, _defaultdict_unflatten, path_entry_type=MappingEntry),  # type: ignore[arg-type]
+    deque: PyTreeNodeRegistryEntry(deque, _deque_flatten, _deque_unflatten, path_entry_type=SequenceEntry),  # type: ignore[arg-type]
+    structseq: PyTreeNodeRegistryEntry(structseq, _structseq_flatten, _structseq_unflatten, path_entry_type=StructSequenceEntry),  # type: ignore[arg-type]
 }
 # pylint: enable=all
 
@@ -495,12 +588,31 @@ _NODETYPE_REGISTRY: dict[type | tuple[str, type], PyTreeNodeRegistryEntry] = {
 def _pytree_node_registry_get(
     cls: type,
     *,
-    namespace: str = __GLOBAL_NAMESPACE,
+    namespace: str = '',
 ) -> PyTreeNodeRegistryEntry | None:
-    handler: PyTreeNodeRegistryEntry | None = _NODETYPE_REGISTRY.get(cls)
-    if handler is not None:
-        return handler
-    handler = _NODETYPE_REGISTRY.get((namespace, cls))
+    handler: PyTreeNodeRegistryEntry | None = None
+    if namespace is not __GLOBAL_NAMESPACE and namespace != '':
+        handler = _NODETYPE_REGISTRY.get((namespace, cls))
+        if handler is not None:
+            return handler
+
+    if _C.is_dict_insertion_ordered(namespace):
+        if cls is dict:
+            return PyTreeNodeRegistryEntry(
+                dict,
+                _dict_insertion_ordered_flatten,  # type: ignore[arg-type]
+                _dict_insertion_ordered_unflatten,  # type: ignore[arg-type]
+                path_entry_type=MappingEntry,
+            )
+        if cls is defaultdict:
+            return PyTreeNodeRegistryEntry(
+                defaultdict,
+                _defaultdict_insertion_ordered_flatten,  # type: ignore[arg-type]
+                _defaultdict_insertion_ordered_unflatten,  # type: ignore[arg-type]
+                path_entry_type=MappingEntry,
+            )
+
+    handler = _NODETYPE_REGISTRY.get(cls)
     if handler is not None:
         return handler
     if is_structseq_class(cls):
@@ -514,196 +626,169 @@ register_pytree_node.get = _pytree_node_registry_get  # type: ignore[attr-define
 del _pytree_node_registry_get
 
 
-class _HashablePartialShim:
-    """Object that delegates :meth:`__call__`, :meth:`__hash__`, and :meth:`__eq__` to another object."""
+@contextlib.contextmanager
+def dict_insertion_ordered(mode: bool, *, namespace: str) -> Generator[None, None, None]:
+    """Context manager to temporarily set the dictionary sorting mode.
 
-    func: Callable[..., Any]
-    args: tuple[Any, ...]
-    keywords: dict[str, Any]
+    This context manager is used to temporarily set the dictionary sorting mode for a specific
+    namespace. The dictionary sorting mode is used to determine whether the keys of a dictionary
+    should be sorted or keeping the insertion order when flattening a pytree.
 
-    def __init__(self, partial_func: functools.partial) -> None:
-        self.partial_func: functools.partial = partial_func
+    >>> tree = {'b': (2, [3, 4]), 'a': 1, 'c': None, 'd': 5}
+    >>> tree_flatten(tree)  # doctest: +IGNORE_WHITESPACE
+    (
+        [1, 2, 3, 4, 5],
+        PyTreeSpec({'a': *, 'b': (*, [*, *]), 'c': None, 'd': *})
+    )
+    >>> with dict_insertion_ordered(True, namespace='some-namespace'):  # doctest: +IGNORE_WHITESPACE
+    ...     tree_flatten(tree, namespace='some-namespace')
+    (
+        [2, 3, 4, 1, 5],
+        PyTreeSpec({'b': (*, [*, *]), 'a': *, 'c': None, 'd': *}, namespace='some-namespace')
+    )
 
-    def __call__(self, *args: Any, **kwargs: Any) -> Any:
-        return self.partial_func(*args, **kwargs)
-
-    def __hash__(self) -> int:
-        return hash(self.partial_func)
-
-    def __eq__(self, other: object) -> bool:
-        if isinstance(other, _HashablePartialShim):
-            return self.partial_func == other.partial_func
-        return self.partial_func == other
-
-
-@register_pytree_node_class(namespace=__GLOBAL_NAMESPACE)
-class Partial(functools.partial, CustomTreeNode[Any]):  # pylint: disable=too-few-public-methods
-    """A version of :func:`functools.partial` that works in pytrees.
-
-    Use it for partial function evaluation in a way that is compatible with transformations,
-    e.g., ``Partial(func, *args, **kwargs)``.
-
-    (You need to explicitly opt-in to this behavior because we did not want to give
-    :func:`functools.partial` different semantics than normal function closures.)
-
-    For example, here is a basic usage of :class:`Partial` in a manner similar to
-    :func:`functools.partial`:
-
-    >>> import operator
-    >>> import torch
-    >>> add_one = Partial(operator.add, torch.ones(()))
-    >>> add_one(torch.tensor([[1, 2], [3, 4]]))
-    tensor([[2., 3.],
-            [4., 5.]])
-
-    Pytree compatibility means that the resulting partial function can be passed as an argument
-    within tree-map functions, which is not possible with a standard :func:`functools.partial`
-    function:
-
-    >>> def call_func_on_cuda(f, *args, **kwargs):
-    ...     f, args, kwargs = tree_map(lambda t: t.cuda(), (f, args, kwargs))
-    ...     return f(*args, **kwargs)
-    ...
-    >>> # doctest: +SKIP
-    >>> tree_map(lambda t: t.cuda(), add_one)
-    Partial(<built-in function add>, tensor(1., device='cuda:0'))
-    >>> call_func_on_cuda(add_one, torch.tensor([[1, 2], [3, 4]]))
-    tensor([[2., 3.],
-            [4., 5.]], device='cuda:0')
-
-    Passing zero arguments to :class:`Partial` effectively wraps the original function, making it a
-    valid argument in tree-map functions:
-
-    >>> # doctest: +SKIP
-    >>> call_func_on_cuda(Partial(torch.add), torch.tensor(1), torch.tensor(2))
-    tensor(3, device='cuda:0')
-
-    Had we passed :func:`operator.add` to ``call_func_on_cuda`` directly, it would have resulted in
-    a :class:`TypeError` or :class:`AttributeError`.
+    Args:
+        mode (bool): The dictionary sorting mode to set.
+        namespace (str): The namespace to set the dictionary sorting mode for.
     """
+    if namespace is not __GLOBAL_NAMESPACE and not isinstance(namespace, str):
+        raise TypeError(f'The namespace must be a string, got {namespace!r}.')
+    if namespace == '':
+        raise ValueError('The namespace cannot be an empty string.')
+    if namespace is __GLOBAL_NAMESPACE:
+        namespace = ''
 
-    func: Callable[..., Any]
-    args: tuple[Any, ...]
-    keywords: dict[str, Any]
+    with __REGISTRY_LOCK:
+        prev = _C.is_dict_insertion_ordered(namespace, inherit_global_namespace=False)
+        _C.set_dict_insertion_ordered(bool(mode), namespace)
 
-    def __new__(cls, func: Callable[..., Any], *args: Any, **keywords: Any) -> Self:
-        """Create a new :class:`Partial` instance."""
-        # In Python 3.10+, if func is itself a functools.partial instance, functools.partial.__new__
-        # would merge the arguments of this Partial instance with the arguments of the func. We box
-        # func in a class that does not (yet) have a `func` attribute to defeat this optimization,
-        # since we care exactly which arguments are considered part of the pytree.
-        if isinstance(func, functools.partial):
-            original_func = func
-            func = _HashablePartialShim(original_func)
-            assert not hasattr(func, 'func'), 'shimmed function should not have a `func` attribute'
-            out = super().__new__(cls, func, *args, **keywords)
-            func.func = original_func.func
-            func.args = original_func.args
-            func.keywords = original_func.keywords
-            return out
-
-        return super().__new__(cls, func, *args, **keywords)
-
-    def tree_flatten(self) -> tuple[tuple[tuple[Any, ...], dict[str, Any]], Callable[..., Any]]:
-        """Flatten the :class:`Partial` instance to children and auxiliary data."""
-        return (self.args, self.keywords), self.func
-
-    @classmethod
-    def tree_unflatten(  # type: ignore[override]
-        cls,
-        metadata: Callable[..., Any],
-        children: tuple[tuple[Any, ...], dict[str, Any]],
-    ) -> Self:
-        """Unflatten the children and auxiliary data into a :class:`Partial` instance."""
-        args, keywords = children
-        return cls(metadata, *args, **keywords)
+    try:
+        yield
+    finally:
+        with __REGISTRY_LOCK:
+            _C.set_dict_insertion_ordered(prev, namespace)
 
 
-class KeyPathEntry(NamedTuple):
-    key: Any
+####################################################################################################
 
-    def __add__(self, other: object) -> KeyPath:
-        if isinstance(other, KeyPathEntry):
-            return KeyPath((self, other))
-        if isinstance(other, KeyPath):
-            return KeyPath((self, *other.keys))
-        return NotImplemented
+with warnings.catch_warnings():
+    warnings.filterwarnings('ignore', category=FutureWarning, module=__name__, append=False)
 
-    def __eq__(self, other: object) -> bool:
-        return isinstance(other, self.__class__) and self.key == other.key
+    @deprecated(
+        'The function `_sorted_keys` is deprecated and will be removed in a future version.',
+        category=FutureWarning,
+    )
+    def _sorted_keys(dct: dict[KT, VT]) -> list[KT]:
+        return total_order_sorted(dct)
 
-    def pprint(self) -> str:
-        """Pretty name of the key path entry."""
-        raise NotImplementedError
+    @deprecated(
+        'The key path API is deprecated and will be removed in a future version. '
+        'Please use the accessor API instead.',
+        category=FutureWarning,
+    )
+    class KeyPathEntry(NamedTuple):
+        key: Any
 
+        def __add__(self, other: object) -> KeyPath:
+            if isinstance(other, KeyPathEntry):
+                return KeyPath((self, other))
+            if isinstance(other, KeyPath):
+                return KeyPath((self, *other.keys))
+            return NotImplemented
 
-class KeyPath(NamedTuple):
-    keys: tuple[KeyPathEntry, ...] = ()
+        def __eq__(self, other: object) -> bool:
+            return isinstance(other, self.__class__) and self.key == other.key
 
-    def __add__(self, other: object) -> KeyPath:
-        if isinstance(other, KeyPathEntry):
-            return KeyPath((*self.keys, other))
-        if isinstance(other, KeyPath):
-            return KeyPath(self.keys + other.keys)
-        return NotImplemented
+        def pprint(self) -> str:
+            """Pretty name of the key path entry."""
+            raise NotImplementedError
 
-    def __eq__(self, other: object) -> bool:
-        return isinstance(other, KeyPath) and self.keys == other.keys
+    @deprecated(
+        'The key path API is deprecated and will be removed in a future version. '
+        'Please use the accessor API instead.',
+        category=FutureWarning,
+    )
+    class KeyPath(NamedTuple):
+        keys: tuple[KeyPathEntry, ...] = ()
 
-    def pprint(self) -> str:
-        """Pretty name of the key path."""
-        if not self.keys:
-            return ' tree root'
-        return ''.join(k.pprint() for k in self.keys)
+        def __add__(self, other: object) -> KeyPath:
+            if isinstance(other, KeyPathEntry):
+                return KeyPath((*self.keys, other))
+            if isinstance(other, KeyPath):
+                return KeyPath(self.keys + other.keys)
+            return NotImplemented
 
+        def __eq__(self, other: object) -> bool:
+            return isinstance(other, KeyPath) and self.keys == other.keys
 
-class GetitemKeyPathEntry(KeyPathEntry):
-    """The key path entry class for sequences and dictionaries."""
+        def pprint(self) -> str:
+            """Pretty name of the key path."""
+            if not self.keys:
+                return ' tree root'
+            return ''.join(k.pprint() for k in self.keys)
 
-    def pprint(self) -> str:
-        """Pretty name of the key path entry."""
-        return f'[{self.key!r}]'
+    @deprecated(
+        'The key path API is deprecated and will be removed in a future version. '
+        'Please use the accessor API instead.',
+        category=FutureWarning,
+    )
+    class GetitemKeyPathEntry(KeyPathEntry):
+        """The key path entry class for sequences and dictionaries."""
 
+        def pprint(self) -> str:
+            """Pretty name of the key path entry."""
+            return f'[{self.key!r}]'
 
-class AttributeKeyPathEntry(KeyPathEntry):
-    """The key path entry class for namedtuples."""
+    @deprecated(
+        'The key path API is deprecated and will be removed in a future version. '
+        'Please use the accessor API instead.',
+        category=FutureWarning,
+    )
+    class AttributeKeyPathEntry(KeyPathEntry):
+        """The key path entry class for namedtuples."""
 
-    def pprint(self) -> str:
-        """Pretty name of the key path entry."""
-        return f'.{self.key}'
+        def pprint(self) -> str:
+            """Pretty name of the key path entry."""
+            return f'.{self.key}'
 
+    @deprecated(
+        'The key path API is deprecated and will be removed in a future version. '
+        'Please use the accessor API instead.',
+        category=FutureWarning,
+    )
+    class FlattenedKeyPathEntry(KeyPathEntry):  # fallback
+        """The fallback key path entry class."""
 
-class FlattenedKeyPathEntry(KeyPathEntry):  # fallback
-    """The fallback key path entry class."""
+        def pprint(self) -> str:
+            """Pretty name of the key path entry."""
+            return f'[<flat index {self.key}>]'
 
-    def pprint(self) -> str:
-        """Pretty name of the key path entry."""
-        return f'[<flat index {self.key}>]'
+    KeyPathHandler = Callable[[PyTree], Sequence[KeyPathEntry]]
+    _KEYPATH_REGISTRY: dict[type[CustomTreeNode], KeyPathHandler] = {}
 
+    @deprecated(
+        'The key path API is deprecated and will be removed in a future version. '
+        'Please use the accessor API instead.',
+        category=FutureWarning,
+    )
+    def register_keypaths(
+        cls: type[CustomTreeNode[T]],
+        handler: KeyPathHandler,
+    ) -> KeyPathHandler:
+        """Register a key path handler for a custom pytree node type."""
+        if not inspect.isclass(cls):
+            raise TypeError(f'Expected a class, got {cls!r}.')
+        if cls in _KEYPATH_REGISTRY:
+            raise ValueError(f'Key path handler for {cls!r} has already been registered.')
 
-KeyPathHandler = Callable[[PyTree], Sequence[KeyPathEntry]]
-_KEYPATH_REGISTRY: dict[type[CustomTreeNode], KeyPathHandler] = {}
+        _KEYPATH_REGISTRY[cls] = handler
+        return handler
 
+    register_keypaths(tuple, lambda tup: list(map(GetitemKeyPathEntry, range(len(tup)))))  # type: ignore[arg-type]
+    register_keypaths(list, lambda lst: list(map(GetitemKeyPathEntry, range(len(lst)))))  # type: ignore[arg-type]
+    register_keypaths(dict, lambda dct: list(map(GetitemKeyPathEntry, _sorted_keys(dct))))  # type: ignore[arg-type]
+    register_keypaths(OrderedDict, lambda odct: list(map(GetitemKeyPathEntry, odct)))  # type: ignore[arg-type,call-overload]
+    register_keypaths(defaultdict, lambda ddct: list(map(GetitemKeyPathEntry, _sorted_keys(ddct))))  # type: ignore[arg-type]
+    register_keypaths(deque, lambda dq: list(map(GetitemKeyPathEntry, range(len(dq)))))  # type: ignore[arg-type]
 
-def register_keypaths(
-    cls: type[CustomTreeNode[T]],
-    handler: KeyPathHandler,
-) -> KeyPathHandler:
-    """Register a key path handler for a custom pytree node type."""
-    if not inspect.isclass(cls):
-        raise TypeError(f'Expected a class, got {cls}.')
-    if cls in _KEYPATH_REGISTRY:
-        raise ValueError(f'Key path handler for {cls} has already been registered.')
-
-    _KEYPATH_REGISTRY[cls] = handler
-    return handler
-
-
-register_keypaths(tuple, lambda tup: list(map(GetitemKeyPathEntry, range(len(tup)))))  # type: ignore[arg-type]
-register_keypaths(list, lambda lst: list(map(GetitemKeyPathEntry, range(len(lst)))))  # type: ignore[arg-type]
-register_keypaths(dict, lambda dct: list(map(GetitemKeyPathEntry, _sorted_keys(dct))))  # type: ignore[arg-type]
-register_keypaths(OrderedDict, lambda odct: list(map(GetitemKeyPathEntry, odct)))  # type: ignore[arg-type,call-overload]
-register_keypaths(defaultdict, lambda ddct: list(map(GetitemKeyPathEntry, _sorted_keys(ddct))))  # type: ignore[arg-type]
-register_keypaths(deque, lambda dq: list(map(GetitemKeyPathEntry, range(len(dq)))))  # type: ignore[arg-type]
-
-register_keypaths.get = _KEYPATH_REGISTRY.get  # type: ignore[attr-defined]
+    register_keypaths.get = _KEYPATH_REGISTRY.get  # type: ignore[attr-defined]
